from collections import OrderedDict, Callable
from .ch import clean_double_diacritic


class DefaultOrderedDict(OrderedDict):
    # Source: http://stackoverflow.com/a/6190500/562769
    def __init__(self, default_factory=None, *a, **kw):
        if (default_factory is not None and
           not isinstance(default_factory, Callable)):
            raise TypeError('first argument must be callable')
        OrderedDict.__init__(self, *a, **kw)
        self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        if self.default_factory is None:
            args = tuple()
        else:
            args = self.default_factory,
        return type(self), args, None, None, iter(self.items())

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        return type(self)(self.default_factory, self)

    def __deepcopy__(self, memo):
        import copy
        return type(self)(self.default_factory,
                          copy.deepcopy(self.items()))

    def __repr__(self):
        return 'OrderedDefaultDict(%s, %s)' % (self.default_factory, OrderedDict.__repr__(self))


def lignesFichier(nf):
    """ L'ensemble de lignes du fichier qui ne sont ni vides ni commentées.

        * Les fichiers de Collatinus ont adopté le point d'exclamation
        * en début de ligne pour introduire un commentaire.
        * Ces lignes doivent être ignorées par le programme.

    :param nf: Nom du fichier
    :type nf: str
    :yield: Ligne de fichier si ce n'est pas un commentaire
    :ytype: str
    """
    with open(nf) as file:
        for line in file.readlines():
            line = line.strip()
            if line and not line.startswith("!") and not line.startswith("! --- "):
                if "!" in line:
                    line, _ = tuple(line.split("!"))  # Suprimer les commentaires
                yield clean_double_diacritic(line)


def flatten(liste):
    return [x for y in liste for x in y]


import re
from .ch import atone
from .util import DefaultOrderedDict


class Desinence(object):
    def __repr__(self):
        return "<pycollatinus.modele.Desinence[{};{};{}]>".format(self.gr(), self.morphoNum(), self.numRad())

    def __init__(self, d, morph, nr, parent=None):
        """ Desinence

        :param d: Graphie avec quantités 
        :type d: str
        :param morph: Numéro de morphologie
        :type morph: int
        :param nr: Numéro de radical accepté par la désinence
        :type nr: int
        :param parent: modèle du lemme et du radical qui utilisent cette désinence
        :type parent: Modele
        """

        # der, dernier caractère de d, s'il est un nombre, le degré de
        # rareté de la désinence, est 10 par défaut.

        self._rarete = 10
        if d and d[-1].isdigit():
            der = int(d[-1])
            self._rarete = der
            d = d[:-1]

        # '-' est la désinence zéro
        if d == "-":
            d = ""
        self._grq = d
        self._gr = atone(self._grq)
        self._morpho = morph
        self._numR = nr
        self._modele = parent

    def gr(self):
        """ Graphie de la désinence, et sans quantités.

        :return: Graphie de la désinence, et sans quantités.
        :rtype: str
        """
        return self._gr

    def grq(self):
        """ Graphie ramiste avec quantités.

        :return: Graphie ramiste avec quantités.
        :rtype: str
        """
        return self._grq

    def modele(self):
        """ Modèle de la désinence.

        :return: Modèle de la désinence.
        :rtype: Modele
        """
        return self._modele

    def rarete(self):
        """ Rareté de la désinence, est 10 par défaut.

        :return: rareté de la désinence, est 10 par défaut.
        :rtype: int
        """
        return self._rarete

    def morphoNum(self):
        """ Numéro de morpho de la désinence.

        :return: Numéro de morpho de la désinence.
        :rtype: int
        """
        return self._morpho

    def numRad(self):
        """ Numéro de radical de la désinence.

        :return: Numéro de radical de la désinence.
        :rtype: int
        """
        return self._numR

    def setModele(self, *m):
        """ Attribue un modèle à la désinence.

        :param m: Modele à attribuer
        :type m: Modele
        """
        self._modele = m


class Modele(object):
    """ Constructeur de la classe modèle.

    :param graphie: Graphie du modele
    :param graphie_accentuee: Graphie accentuée du modèle
    :param parent: Lemmatiseur
    :type parent: pycollatinus.lemmatiseur.Lemmatiseur
    :param pere: Modele from which the modele derive
    :type pere: Modele
    """
    RE = re.compile("[:;]*([\\w]*)\\+{0,1}(\\$\\w+)")

    def __repr__(self):
        return "<pycollatinus.modele.Modele[{}]>".format(self.gr())

    def __init__(self, graphie: str, graphie_accentuee: str, parent=None, pere=None):
        self._lemmatiseur = parent
        self._pere = pere
        self._desinences = DefaultOrderedDict(list)
        self.msuff = DefaultOrderedDict(list)  # list of int
        self._absents = []  # List of int
        self._genRadicaux = {}  # int -> str
        self._gr = graphie
        self._grq = graphie_accentuee

    def hasRadical(self, numRad):
        """ Vérifie que l'objet à le radical donné

        :param numRad: Identifiant de radical
        :type numRad: int
        :return: Si l'objet à la radical
        :rtype: bool
        """
        return numRad in self._genRadicaux

    def absent(self, a):
        """ Renvoie True si la morpho de rang a n'existe pas dans le modèle. 

        Certains modèles, exemple, n'ont pas de singulier, certains verbes n'ont pas de passif.

        :return: Si la morpho de rang a n'existe pas dans le modèle
        :rtype: bool
        """
        return a in self._absents

    def absents(self):
        """ Retourne la liste des numéros des morphos absentes.

        :return: Liste des numéros des morphos absentes.
        :rtype: list of int
        """
        return self._absents

    def cles_radicaux(self):
        """ Liste des numéros de radicaux utilisés, et rangés dans la map _genRadicaux.

        :return: Liste des numéros de radicaux utilisés, et rangés dans la map _genRadicaux.
        :rtype: list of int
        """
        return sorted(list(self._genRadicaux.keys()))

    def clone(self, d):
        """ Crée une Désinence copiée sur la désinence d.

        :param d: Desinence à copier
        :type d: Desinence 
        :return: Désinence copiée sur la désinence d.
        :rtype: Desinence
        """
        return Desinence(d.grq(), d.morphoNum(), d.numRad(), self)

    def deja(self, m):
        """ Renvoie True si la désinence a une morpho de rang m. 

        Permet de savoir s'il faut aller chercher la désinence de morpho m chez le modèle père.

        :return: Si la désinence a une morpho de rang m. 
        :rtype: bool
        """
        return m in self._desinences

    def desinences(self, d=None):
        """ Renvoie la liste des désinence de morpho d du modèle.

        :param d: Identifant de morphologie
        :type d: int
        :return: Liste des désinence de morpho d du modèle ou toutes les désinences du modèle (Applaties en python, doute sur original)
        :rtype: list of Desinence
        """
        if d:
            return self._desinences[d]
        return [x for morpho_values in self._desinences.values() for x in morpho_values]

    def estUn(self, m):
        """ Renvoie True si le modèle se nomme m, si l'un de ses ancêtre se nomme m

        :param m: Nom de modele
        :type m: str
        :return: si le modèle se nomme m, si l'un de ses ancêtre se nomme m
        :rtype: bool
        """
        if self._gr == m:
            return True
        if self._pere is not None:
            return self._pere.estUn(m)
        return False

    def gr(self):
        """ Nom du modèle

        :return: Nom du modèle.
        :rtype: str
        """
        return self._gr

    def genRadical(self, r):
        """ Chaîne permettant de calculer un radical à partir de la forme canonique d'un lemme. 

        :param r: Numéro du radical
        :type r: int
        :return: Chaîne permettant de calculer un radical à partir de la forme canonique d'un lemme. 
        :rtype: str
        """
        return self._genRadicaux[r]

    def morphos(self):
        """ Liste des numéros des désinences définies par le modèle.

        :return: Liste des numéros des désinences définies par le modèle.
        :rtype: list of int
        """
        return list(self._desinences.keys())

    def pos(self):
        """ Retourne la catégorie du modèle, utilisant les ancêtres du modèle.

        :return: Catégorie du modèle
        :rtype: str
        """
        if self.estUn("uita") or self.estUn("lupus") or self.estUn("miles") or self.estUn("manus") or self.estUn("res") or self.estUn("perseus"):
            return 'n'
        if self.estUn("doctus") or self.estUn("fortis"):
            return 'a'
        if self.estUn("amo") or self.estUn("imitor"):
            return 'v'
        return 'd'



from .ch import estRomain, deramise
from .util import DefaultOrderedDict
from .lemme import Lemme
from .modele import Modele
from .parser import Parser
import os
import re
from pickle import dump, load


SPACES = re.compile("\W")


class Lemmatiseur(object):
    """ Main lemmatiseur object copied directly from CPP

    :ivar _radicaux: Dictionary of Radicaux
    :type _radicaux: dict[str, list[pycollatinus.rad.Radical]]
    """
    def __init__(self, load=True):
        """"""
        self._resDir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
        self._cible = "fr"  # Langue cible
        self._modeles = {}  # Modeles
        self._cibles = {}
        self._lemmes = {}  # Lemmes

        self._assims = {}  # str -> str
        self._assimsq = {}  # str -> str
        self._contractions = {}  # str -> str
        self._variables = {}  # str -> str # Where key starts with $

        self._radicaux = DefaultOrderedDict(list)  # List of Radicaux
        self._desinences = DefaultOrderedDict(list)  # List of Desinence
        self._irregs = DefaultOrderedDict(list)  # List of Irreg
        self._morphos = {"fr": {}}  # List of Strings
        self._cas = DefaultOrderedDict(list)  # List of Strings
        self._genres = DefaultOrderedDict(list)  # List of Strings
        self._nombres = DefaultOrderedDict(list)  # List of Strings
        self._temps = DefaultOrderedDict(list)  # List of Strings
        self._modes = DefaultOrderedDict(list)  # List of Strings
        self._voix = DefaultOrderedDict(list)  # List of Strings
        self._motsClefs = DefaultOrderedDict(list)  # List of Strings

        if load is True:
            Parser(self, path=self._resDir).parse()

    def compile(self):
        """ Compile le lemmatiseur localement
        """
        with open(self.path("compiled.pickle"), "wb") as file:
            dump(self, file)

    @staticmethod
    def load(path=None):
        """ Compile le lemmatiseur localement
        """
        if path is None:
            path = os.path.join(os.path.join(os.path.dirname(os.path.abspath(__file__)), "data"))
            path = os.path.join(path, "compiled.pickle")
        with open(path, "rb") as file:
            return load(file)

    def path(self, nf):
        """ Compute the path for the file to load

        :rtype: str
        """
        return os.path.join(self._resDir, nf)

    def assims(self, mot):
        """ Cherche si la chaîne a peut subir une assimilation, renvoie cette chaîne éventuellement assimilée.

        :param mot: Mot pour lequel on doit vérifier des assimilations
        :type mot: str
        :return: Mot assimilé
        :rtype: str
        """
        for replaced, replacement in self._assimsq.items():
            if mot.startswith(replaced):
                mot = mot.replace(replaced, replacement)
                return mot
        return mot

    def desassims(self, mot):
        """ Cherche si la chaîne a peut subir une assimilation inversée, renvoie cette chaîne éventuellement assimilée.

        :param mot: Mot pour lequel on doit vérifier des assimilations
        :type mot: str
        :return: Mot assimilé
        :rtype: str
        """
        for replacement, replaced in self._assimsq.items():
            if mot.startswith(replaced):
                mot = mot.replace(replaced, replacement)
                return mot
        return mot

    def modele(self, m):
        """ Retrouve le modele pour la clef m

        :param m: Nom du modele
        :type m: str
        :return: Modele trouvé
        :rtype: Modele
        """
        return self._modeles[m]

    def lemme(self, lemme):
        """ Retrouve le lemme pour la clef lemme

        :param lemme: Nom du lemme
        :type lemme: str
        :return: Lemme trouvé
        :rtype: Lemme
        """
        return self._lemmes[lemme]

    def morpho(self, m):
        """ Renvoie la chaîne de rang m dans la liste des morphologies donnée par le fichier data/morphos.la

        :param m: Indice de morphologie
        :type m: int
        :return: Chaîne de rang m dans la liste des morphologies donnée par le fichier data/morphos.la
        :rtype: str
        """
        l = "fr"

        # TODO : Si ajout langue de traduction, il faudra convertir ce qui suit
        ###################
        # if self._morphos.keys().contains(_cible.mid(0,2))) l = _cible.mid(0,2:
        # elif (_cible.size() > 4) and (_morphos.keys().contains(_cible.mid(3,2))):
        #    l = _cible.mid(3,2)
        ####################

        if m < 0 or m > len(self._morphos[l]):
            raise KeyError("Morphology %s requested but not found" % m)
        if m == len(self._morphos[l]):
            return "-"
        return self._morphos[l][m]

    @staticmethod
    def format_result(form, lemma, morphos=None, with_pos=False, raw_obj=False):
        r = {"form": form, "lemma": lemma.gr(), "morph": morphos or ""}
        if raw_obj:
            r["lemma"] = lemma
        if with_pos:
            r["pos"] = lemma.pos()
        return r

    def lemmatise_multiple(self, string, pos=False, get_lemma_object=False, as_list=True):
        """ Lemmatise une liste complète

        :param string: Chaîne à lemmatiser
        :param pos: Récupère la POS
        :param get_lemma_object: Retrieve Lemma object instead of string representation of lemma
        :param as_list: Retrieve a list of generators instead of a list if set to false
        """
        mots = SPACES.split(string)
        resultats = [self.lemmatise(mot, pos=pos, get_lemma_object=get_lemma_object) for mot in mots]
        if as_list:
            resultats = [list(r) for r in resultats]
        return resultats

    def _lemmatise_assims(self, f, *args, **kwargs):
        """ Lemmatise un mot f avec son assimilation

        :param f: Mot à lemmatiser
        :param pos: Récupère la POS
        :param get_lemma_object: Retrieve Lemma object instead of string representation of lemma
        :param results: Current results
        """
        forme_assimilee = self.assims(f)
        if forme_assimilee != f:
            for proposal in self._lemmatise(forme_assimilee, *args, **kwargs):
                yield proposal

    def _lemmatise_roman_numerals(self, form, pos=False, get_lemma_object=False):
        """ Lemmatise un mot f si c'est un nombre romain

        :param form: Mot à lemmatiser
        :param pos: Récupère la POS
        :param get_lemma_object: Retrieve Lemma object instead of string representation of lemma
        """
        if estRomain(form):
            _lemma = Lemme(
                cle=form, graphie_accentuee=form, graphie=form, parent=self, origin=0, pos="a",
                modele=self.modele("inv")
            )
            yield Lemmatiseur.format_result(
                form=form,
                lemma=_lemma,
                with_pos=pos,
                raw_obj=get_lemma_object
            )

        if form.upper() != form:
            yield from self._lemmatise_roman_numerals(form.upper(), pos=pos, get_lemma_object=get_lemma_object)

    def _lemmatise_contractions(self, f, *args, **kwargs):
        """ Lemmatise un mot f avec sa contraction

        :param f: Mot à lemmatiser
        :yield: Match formated like in _lemmatise()
        """
        fd = f
        for contraction, decontraction in self._contractions.items():
            if fd.endswith(contraction):
                fd = f[:-len(contraction)]
                if "v" in fd or "V" in fd:
                    fd += decontraction
                else:
                    fd += deramise(decontraction)
                yield from self._lemmatise(fd, *args, **kwargs)

    def _lemmatise_desassims(self, f, *args, **kwargs):
        """ Lemmatise un mot f avec sa désassimilation

        :param f: Mot à lemmatiser
        :yield: Match formated like in _lemmatise()
        """
        forme_assimilee = self.desassims(f)
        if forme_assimilee != f:
            for proposal in self._lemmatise(forme_assimilee, *args, **kwargs):
                yield proposal

    def lemmatise(self, f, pos=False, get_lemma_object=False, lower=True):
        """ Lemmatise un mot f

        :param f: Mot à lemmatiser
        :param pos: Récupère la POS
        :param get_lemma_object: Retrieve Lemma object instead of string representation of lemma
        :param lower: Need to check lowercase version
        """
        if lower:
            # We do not run numeral on lower
            yield from self._lemmatise_roman_numerals(f, pos=pos, get_lemma_object=get_lemma_object)

            # We run on the lower version
            if f.lower() != f:
                yield from self.lemmatise(f.lower(), pos=pos, get_lemma_object=get_lemma_object, lower=False)

        f = deramise(f)
        yield from self._lemmatise(f, pos=pos, get_lemma_object=get_lemma_object)
        yield from self._lemmatise_assims(f, pos=pos, get_lemma_object=get_lemma_object)
        yield from self._lemmatise_desassims(f, pos=pos, get_lemma_object=get_lemma_object)
        yield from self._lemmatise_contractions(f, pos=pos, get_lemma_object=get_lemma_object)

    def _lemmatise(self, form, pos=False, get_lemma_object=False):
        """ Lemmatise un mot f

        :param f: Mot à lemmatiser
        :param pos: Récupère la POS
        :param get_lemma_object: Retrieve Lemma object instead of string representation of lemma
        :param check_assims: Vérifie les assimilations.
        """
        result = []
        if not form:
            return result

        # formes irrégulières
        for irr in self._irregs[form]:
            for m in irr.morphos():
                yield Lemmatiseur.format_result(
                    form=form, lemma=irr, morphos=self.morpho(m), with_pos=pos,
                    raw_obj=get_lemma_object
                )

        # radical + désinence
        for i in range(len(form)+1):
            if i == 0:
                radical = ""
                desinence = form
            else:
                radical = form[:i]
                desinence = form[i:]
            ldes = self._desinences.get(desinence, None)  # List of desinences
            if ldes is None:
                continue
            # Je regarde d'abord si d est une désinence possible,
            # car il y a moins de désinences que de radicaux.
            # Je fais la recherche sur les radicaux seulement si la désinence existe.
            lrad = [] + self._radicaux.get(radical, [])

            # ii noté ī
            # 1. Patauium, gén. Pataui : Patau.i . Patau+i.i
            # 2. conubium, conubis : conubi.s . conubi.i+s
            if desinence.startswith('i') and not desinence.startswith("ii") and not radical.endswith('i'):
                lrad += self._radicaux.get(radical + "i", [])

            if len(lrad) == 0:  # Il n'y a rien à faire si le radical n'existe pas.
                continue

            for rad in lrad:
                lemme = rad.lemme()
                for des in ldes:
                    if des.modele() == lemme.modele() and des.numRad() == rad.numRad() and not lemme.estIrregExcl(des.morphoNum()):
                        # Commented this part because we are not using quantity right now.
                        yield Lemmatiseur.format_result(
                                form, lemme, morphos=self.morpho(des.morphoNum()), with_pos=pos,
                                raw_obj=get_lemma_object
                            )

        return result

    def variable(self, v):
        """ Permet de remplacer la métavariable v
            par son contenu. Ces métavariables sont
            utilisées par le fichier modeles.la, pour
            éviter de répéter des suites de désinences.
            Elles sont repérées comme en PHP, leur
            premier caractère $.

            :param v: Nom de la variable
            :type v: str
            :return: Valeur de variable
            :rtype: str
        """
        return self._variables[v]



from .ch import atone, communes
import re
from .util import DefaultOrderedDict
from .modele import Modele


class Radical(object):
    def __init__(self, g, n, parent=None):
        """ Représentation d'un radical

        :param g: Forme canonique du radical avec quantité
        :type g: str
        :param n: Numéro du radical
        :type n: int
        """
        self._lemme = parent
        self._grq = communes(g)
        self._gr = atone(g)
        self._numero = n

    def gr(self):
        """ Renvoie la graphie du radical dépourvue de diacritiques.

         :return: Graphie du radical dépourvue de diacritiques.
        :rtype: str
        """
        return self._gr

    def grq(self):
        """ Renvoie la graphie du radical pourvue de ѕes diacritiques.

        :return: raphie du radical pourvue de ѕes diacritiques.
        :rtype: str
        """
        return self._grq

    def lemme(self):
        """ Le lemme auquel appartient le radical.

        :return: Le lemme auquel appartient le radical.
        :rtype: Lemme
        """
        return self._lemme

    def set_lemme(self, lemme):
        """ Set the parent lemme

        :param lemme: Lemme
        :type lemme: Lemme
        """

        self._lemme = lemme

    def modele(self):
        """ Le modèle de flexion du radical

        :return: Le modèle de flexion du radical
        :rtype: str
        """
        return self._lemme.modele()

    def numRad(self):
        """ Le numéro du radical.

        :return: Le numéro du radical.
        :rtype: int
        """
        return self._numero

    def __repr__(self):
        return "<pycollatinus.lemme.Radical[{}:{}]>".format(self.gr(), self.numRad())


class Lemme(object):
    RENVOI = re.compile("cf\\.\\s(\\w+)$")

    def __repr__(self):
        return "<pycollatinus.lemme.Lemme[{}:modele-{}]>".format(self.cle(), self.grModele())

    def __eq__(self, other):
        return self.__repr__() == other.__repr__()

    def __init__(self,
                 cle: str, graphie: str, graphie_accentuee: str,
                 modele: Modele, parent,
                 radicaux: DefaultOrderedDict=None, origin: int =0, pos: str="-",
                 nombre_homonymie: int=0, nbOcc: int=0):
        """  Generate a lemma object

        :param cle: Unique Identifier of the Lemma
        :type cle: str
        :param graphie: Graphie
        :type graphie: str
        :param graphie_accentuee: Accentuated Graphie
        :type graphie_accentuee: str
        :param modele: Modele of the Lemma
        :type modele: pycollatinus.modele.Modele
        :param parent: Lemmatiseur
        :type parent: pycollatinus.Lemmatiseur
        :param radicaux: Dictionary of list of radicaux
        :type radicaux: DefaultOrderedDict(list)
        :param origin: Origin of the lemma (0 curated, 1 automatic import)
        :type origin: int
        :param pos: POS tag
        :type pos: str
        :param nombre_homonymie: Number to distinguish homonyms
        :type nombre_homonymie: int
        :param nbOcc: Number of known occurences
        :type nbOcc: int
        """
        self._lemmatiseur = parent
        self._radicaux = radicaux or DefaultOrderedDict(list)
        self._irregs = []  # list of Irreg
        self._morphosIrrExcl = []  # list of int
        self._nh = nombre_homonymie
        self._nbOcc = nbOcc
        self._cle = cle
        self._grq = graphie_accentuee
        self._gr = graphie
        self._grModele = modele.gr()
        self._modele = modele
        self._indMorph = ""
        self._renvoi = None
        self._origin = origin
        self._pos = pos

    def ajIrreg(self, irr):
        """ Ajoute au lemme l'obet irr, représente
                 une forme irrégulière. Lorsque les formes irrégulières
                 sont trop nombreuses, lorsque plusieurs lemmes
                 ont des formes analogues, vaut ajouter un modèle
                 dans data/modeles.la.

        :param irr: Irrégulier
        :type irr: pycollatinus.irregs.Irreg
        """
        self._irregs.append(irr)
        # ajouter les numéros de morpho à la liste
        # des morphos irrégulières du lemme :
        if irr.exclusif():
            self._morphosIrrExcl += irr.morphos()

    def ajRadical(self, i, r=None):
        """ Ajoute le radical r de numéro i à la map des radicaux du lemme.

        :param i: Index de radical
        :type i: int
        :param r: Radical à ajouter
        :type r: Radical
        """
        if r:
            self._radicaux[i].append(r)

    def cle(self):
        """ Renvoie la clé sous laquel le lemme est enregistré dans le lemmatiseur parent.

        :return:
        :rtype:
        """
        return self._cle

    def cles_radicaux(self):
        """ Retourne toutes les clés (formes non-ramistes sans diacritiques) de la map des radicaux du lemme.
        """
        return self._radicaux.keys()

    def estIrregExcl(self, nm):
        """ Renvoie vrai si la forme irrégulière avec le n° nm remplace celle construite sur le radical , si la forme régulière existe aussi.

        :param nm: Numéro de morpho
        :type nm: int
        :return: Statut de forme irrégulière
        :rtype: bool
        """
        return nm in self._morphosIrrExcl

    def genre(self):
        """ Cette routine convertit les indications morphologiques, données dans le fichier lemmes.la, pour exprimer le genre du mot dans la langue courante.

        :return: Genre
        :rtype: str
        """
        _genre = ""
        if " m." in self._indMorph:
            _genre += "m"
        if " f." in self._indMorph:
            _genre += "f"
        if " n." in self._indMorph:
            _genre += "n"
        _genre = _genre.strip()
        if self._renvoi and not _genre:
            lr = self._lemmatiseur.lemme(self._renvoi)
            if lr:
                return lr.genre()

        return _genre

    def gr(self):
        """ Retourne la graphie ramiste du lemme sans diacritiques.

        :return: Graphie ramiste du lemme sans diacritiques.
        :rtype: str
        """
        return self._gr

    def grq(self):
        """ Retourne la graphie ramiste du lemme sans diacritiques.

        :return: Graphie ramiste du lemme sans diacritiques.
        :rtype: str
        """
        return self._grq

    def grModele(self):
        """ Retourne la graphie du modèle du lemme.

        :return: Graphie du modèle du lemme.
        :rtype: ???
        """
        return self._grModele

    def indMorph(self):
        """ Returne l'index de morphologie
        """
        return self._indMorph

    def irreg(self, i):
        """ Renvoie la forme irrégulière de morpho i. excl devient True si elle est exclusive, sinon.

        :return: Forme irrégulière de morpho i, Exclusivité
        :rtype: tuple.<str, bool>
        """
        excl = False
        for ir in self._irregs:
            if i in ir.morphos():
                return ir.grq(), ir.exclusif
        return "", excl

    def modele(self):
        """ Renvoie l'objet modèle du lemme.

        :return: Modèle du lemme
        :rtype: pycollatinus.modele.Modele
        """
        return self._modele

    def nbOcc(self):
        """ Renvoie le nombre d'occurrences du lemme dans les textes du LASLA.

        :return: Nombre d'occurrences du lemme dans les textes du LASLA.
        :rtype: int
        """
        return self._nbOcc

    def clearOcc(self):
        """ Initialise le nombre d'occurrences. """
        self._nbOcc = 1

    def nh(self):
        """ Renvoie le numéro d'homonymie du lemme.

        :return: Numéro d'homonymie du lemme.
        :rtype: int
        """
        return self._nh

    def origin(self):
        """ Renvoie l'origine du lemme : 0 pour le lexique de base, pour l'extension. 

        :return: Origine du lemme : 0 pour le lexique de base, pour l'extension. 
        :rtype: int
        """
        return self._origin

    def pos(self):
        """ Renvoie un caractère représentant la catégorie (part of speech, orationis) du lemme.

        :return: Caractère représentant la catégorie (part of speech, orationis) du lemme.
        :rtype: str
        """
        if not self._pos and self._renvoi:
            lr = self._lemmatiseur.lemme(self._renvoi)
            if lr:
                return lr.pos()
        return self._pos

    def radical(self, r):
        """  Renvoie le radical numéro r du lemme.
        
        :param r: Numéro de radical
        :type r: int
        :return: Radicaux enregistré au numéro R
        :rtype: list of Radical
        """
        return self._radicaux[r]

    def renvoi(self):
        """ Renvoie True si le lemme est une forme alternative renvoyant à une autre entrée du lexique.

        :return: Statut forme alternative renvoyant à une autre entrée du lexique
        :rtype: bool
        """
        return "cf. " in self._indMorph

    def traduction(self, l):
        """ Renvoie la traduction du lemme dans la langue cible l (2 caractères, plus
         pour donner l'ordre des langues de secours). J'ai opté pour un format "l1.l2.l3" où les trois langues sont en 2 caractères.
        """
        pass

    def __lt__(self, l):
        """ Vrai si la fréquence du lemme de gauche est inférieure à celle de celui de droite. 

        commenté : vrai si la graphie du lemme de gauche précède celle de celui de droite dans l'ordre alphabétique.

        :param l: Other lemma
        :type l: Lemme
        :return: Fréquence du lemme de gauche est inférieure à celle de celui de droite. 
        :rtype: Bool
        """
        return self._nbOcc < l.nbOcc()

    def setHyphen(self, h):
        """ Stocke l'information sur la césure étymologique du lemme

        :param h: indique où se fait la césure. 
        :type h: int
        """
        self._hyphen = h

    def getHyphen(self):
        """ Retourne la césure étymologique du lemme

        :return: Césure étymologique du lemme
        """
        return self._hyphen

    def possible_forms(self):
        """ Generate a list of possible forms for the current lemma

        :returns: List of possible forms for the current lemma
        :rtype: [str]
        """
        forms = []
        for morph in self.modele().morphos():
            for desinence in self.modele().desinences(morph):
                radicaux = self.radical(desinence.numRad())
                if isinstance(radicaux, Radical):
                    forms.append(radicaux.gr() + desinence.gr())
                else:
                    for rad in radicaux:
                        forms.append(rad.gr() + desinence.gr())
        return list(set(forms))



from .ch import deramise, atone, listeI, allonge, simplified
from .util import lignesFichier, flatten, DefaultOrderedDict
from .lemme import Lemme, Radical
from .irregs import Irreg
from .modele import Modele, Desinence
from .error import UnknownModeleConfigurationKey, MissingRadical
import os
import warnings
import re


SPACES = re.compile("\W")

MODELE_CLEFS = [
    "pere", # 0
    "des",  # 1
    "des+",  # 2
    "R",  # 3
    "abs",  # 4
    "suf",  # 5
    "sufd",  # 6
    "abs+",  # 7
]


class Parser(object):
    """ Parser object that fills the Lemmatiseur

    :param lemmatiseur: Lemmatiseur object to fill
    :param path: Path in which we find Collatinus data
    :param cible: Language for morphology
    """
    def __init__(self, lemmatiseur, path=None, cible="fr"):
        """"""
        self.__lemmatiseur__ = lemmatiseur
        self.__data_directory__ = path or os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
        self.__cible__ = cible

    @property
    def data_directory(self):
        return self.__data_directory__

    @property
    def lemmatiseur(self):
        return self.__lemmatiseur__

    def parse(self):
        self.ajAssims()
        self.ajContractions()
        self.ajMorphos(self.__cible__)  # Note : from lisModeles
        self.ajModeles()  # Note : from lisModeles
        self.ajLexiques()  # Note : from lisLexique
        self.ajExtensions()  # Note : from lisLexique
        self.ajIrreguliers()

    def path(self, nf):
        """ Compute the path for the file to load

        :rtype: str
        """
        return os.path.join(self.data_directory, nf)

    def ajMorphos(self, lang="fr"):
        for ligne in lignesFichier(self.path("morphos." + lang)):
            if ":" not in ligne:
                break
            morph_id, morph_str = tuple(ligne.split(":"))
            self.lemmatiseur._morphos[lang][int(morph_id)] = morph_str

    def ajIrreguliers(self):
        """ Chargement des formes irrégulières du fichier data/irregs.la
        """
        lignes = lignesFichier(self.path("irregs.la"))
        for lin in lignes:
            try:
                irr = self.parse_irreg(lin)
                self.lemmatiseur._irregs[deramise(irr.gr())].append(irr)
            except Exception as E:
                warnings.warn("Erreur au chargement de l'irrégulier\n" + lin + "\n" + str(E))
                raise E

        for irr in flatten(self.lemmatiseur._irregs.values()):
            irr.lemme().ajIrreg(irr)

    def ajAssims(self):
        """ Charge et définit les débuts de mots non-assimilés, associe à chacun sa forme assimilée.
        """
        for lin in lignesFichier(self.path("assimilations.la")):
            ass1, ass2 = tuple(lin.split(':'))
            self.lemmatiseur._assims[ass1] = ass2
            self.lemmatiseur._assimsq[atone(ass1)] = atone(ass2)

    def ajContractions(self):
        """ Charge et établit une liste qui donne, chaque contraction, forme non contracte qui lui correspond.
        """
        for lin in lignesFichier(self.path("contractions.la")):
            ass1, ass2 = tuple(lin.split(':'))
            self.lemmatiseur._contractions[ass1] = ass2

    def lisFichierLexique(self, filepath):
        """ Lecture des lemmes, et enregistrement de leurs radicaux

        :param filepath: Chemin du fichier à charger
        :type filepath: str
        """
        orig = int(filepath.endswith("ext.la"))
        lignes = lignesFichier(filepath)
        for ligne in lignes:
            self.parse_lemme(ligne, orig)

    def parse_lemme(self, linea: str, origin: int=0, _deramise: bool=True):
        """ Constructeur de la classe Lemme à partir de la ligne linea.

        Exemple de linea avec numéro d'éclat:
            # cădo|lego|cĕcĭd|cās|is, ere, cecidi, casum|687
            #   0 | 1  | 2   | 3 |     4                | 5

        :param linea: Ligne à parser
        :type linea: str
        :param origin: 0 for original curated lemma, 1 for automatic import from Gaffiot
        :type origin: int
        :param _deramise: Force the deramisation of the normalized graphie
        :type _deramise: bool
        """
        eclats = linea.split('|')
        lg = eclats[0].split('=')

        if _deramise:
            cle = atone(deramise(lg[0]))
        else:
            cle = atone(lg[0])

        # Some lemma have homonyms, we have a number do differentiate them
        nh = 0
        if cle[-1].isnumeric():
            nh = int(cle[-1])
            grd = cle[:-1]
        else:
            grd = cle

        # We setup the accentuated graphie
        if len(lg) == 1:
            grq = grd
        else:
            grq = lg[1]
        # Pour l'affichage des dictionnaires, élimine les doubles de la forme canonique
        gr = atone(grq.split(",")[0])
        grModele = eclats[1]
        modele = self.lemmatiseur.modele(grModele)

        # contrôle de format. la liste doit avoir 6 items
        if len(eclats) < 6:
            warnings.warn("Ligne mal formée : " + gr + "\n ---Dernier champ " + eclats[-1] + "\n ---" + linea)

        radicaux = DefaultOrderedDict(list)
        # lecture des radicaux, 2 et 3
        for i in range(2, 4):
            if eclats[i]:
                lrad = eclats[i].split(',')
                for rad in lrad:
                    radicaux[i - 1].append(Radical(rad, i - 1))

        # Gros doute sur le fonctionnement ici
        indMorph = eclats[4]
        match_renvoi = Lemme.RENVOI.match(indMorph)
        if match_renvoi is not None:
            renvoi = match_renvoi.group(1)
        else:
            renvoi = ""

        pos = ""
        if "adj." in indMorph:
            pos += 'a'
        if "conj" in indMorph:
            pos += 'c'
        if "excl." in indMorph:
            pos += 'e'
        if "interj" in indMorph:
            pos += 'i'
        if "num." in indMorph:
            pos += 'm'
        if "pron." in indMorph:
            pos += 'p'
        if "prép" in indMorph:
            pos += 'r'
        if "adv" in indMorph:
            pos += 'd'
        if " nom " in indMorph:
            pos += 'n'
        if "npr." in indMorph:
            pos += 'n'
        if not pos:
            pos = modele.pos()  # Je prends le POS du modèle
            if pos == "d" and renvoi:
                pos = ""
                # S'il y a un renvoi (cf.) et que le modèle a donné le POS "d" (adverbe),
                # je prendrai le pos du renvoi (les indéclinables ont le POS par défaut "d").
                # Je ne peux pas le faire maintenant !

        # Nombre d'occurrences
        if len(eclats[5]):
            nbOcc = int(eclats[5])
        else:
            nbOcc = 1

        lemma = Lemme(
            cle=cle,
            graphie=gr, graphie_accentuee=grq,
            modele=modele, radicaux=radicaux,
            parent=self.lemmatiseur,
            nombre_homonymie=nh, nbOcc=nbOcc,
            origin=origin, pos=pos
        )
        # We register the lemma for each radical
        for radNum in lemma._radicaux:
            for rad in lemma._radicaux[radNum]:
                rad.set_lemme(lemma)
        self._register_lemme(lemma)
        return lemma

    def register_modele(self, modele: Modele):
        """ Register a modele onto the lemmatizer

        :param modele: Modele to register
        """
        self.lemmatiseur._modeles[modele.gr()] = modele

    def parse_modele(self, list_of_lines: list):
        """ Chaque item de la liste list_of_lines est constitué de champs séparé par le caractère ':'. Le premier champ est un mot clé. Le parent est le lemmatiseur.
        Pour le format du fichier data/modeles.la, la documentation utilisateur.

        :param list_of_lines: Liste de champs séparés par le caractère ":". Le premier champ est un mot-clef
        :return: model
        :rtype: Modele
        """
        grq = list_of_lines[0].split(":")[1]
        gr = atone(grq)
        modele = Modele(graphie=gr, graphie_accentuee=grq, parent=self.lemmatiseur)
        pere = None

        for original_l in list_of_lines[1:]:
            # remplacement des variables par leur valeur
            l = "" + original_l

            # TODO : Ajouté * pour la premiere capture pour rendre le prefix optionnel...
            for pre, v in Modele.RE.findall(l):
                var = self.lemmatiseur.variable(v)
                if pre:
                    var = var.replace(";", ";" + pre)
                l = l.replace(v, var)

            eclats = simplified(l).split(":")
            # type_of_line -> pere des  des+  R   abs
            #                  0    1    2    3    4
            type_of_line = MODELE_CLEFS.index(eclats[0])
            if type_of_line == 0:  # père
                modele._pere = pere = self.lemmatiseur.modele(eclats[1])
            elif type_of_line == 1 or type_of_line == 2:
                # des+: désinences s'ajoutant à celles du père
                # des: désinences écrasant celles du père
                index_morphologies = listeI(eclats[1])  # Anciennement li
                radical = int(eclats[2])  # Anciennement "r"
                liste_desinences = eclats[3].split(';')  # Anciennement "ld"
                for i in range(len(index_morphologies)):
                    if i < len(liste_desinences):
                        liste_desinences_definitives = liste_desinences[i].split(',')  # Anciennement "ldd"
                    else:
                        liste_desinences_definitives = liste_desinences[-1].split(',')

                    for graphie in liste_desinences_definitives:
                        nd = Desinence(graphie, index_morphologies[i], nr=radical, parent=modele)
                        modele._desinences[nd.morphoNum()].append(nd)
                        self.ajDesinence(nd)

                # si des+, chercher les autres désinences chez le père :
                if type_of_line == 2:
                    for i in index_morphologies:
                        for dp in pere.desinences(i):
                            dh = modele.clone(dp)  # cloner la désinence
                            modele._desinences[i].append(dh)
                            self.ajDesinence(dh)

            elif type_of_line == 3:  # R:n: radical n
                modele._genRadicaux[int(eclats[1])] = eclats[2]
            elif type_of_line == 4:  # abs
                modele._absents = listeI(eclats[1])
            elif type_of_line == 7:  # abs+
                modele._absents += listeI(eclats[1])
            elif type_of_line == 5:  # suffixes suf:<intervalle>:valeur
                lsuf = listeI(eclats[1])
                gr = eclats[2]  # TODO verif : bien formée ?
                for m in lsuf:
                    modele.msuff[gr].append(m)
            elif type_of_line == 6:  # sufd: les désinences du père, suffixées
                if pere:
                    suf = eclats[1]
                    ld = pere.desinences()
                    for d in ld:
                        if d.morphoNum() in modele._absents:
                            continue
                        nd = allonge(d.grq())
                        dsuf = Desinence(nd+suf, d.morphoNum(), d.numRad(), modele)
                        modele._desinences[dsuf.morphoNum()].append(dsuf)
                        self.ajDesinence(dsuf)
            else:
                warnings.warn("Modele : Erreur pour " + l, UnknownModeleConfigurationKey)

        # père
        if pere:

            for m in pere.morphos():
                # héritage des désinence
                if modele.deja(m):
                    continue
                ld = pere.desinences(m)
                for d in ld:
                    if d.morphoNum() in modele._absents:  # morpho absente chez le descendant
                        continue
                    dh = modele.clone(d)
                    modele._desinences[dh.morphoNum()].append(dh)
                    self.ajDesinence(dh)

            # héritage des radicaux
            for numRad in set([d.numRad() for d in flatten(modele._desinences.values())]):
                if numRad not in modele._genRadicaux:
                    if pere.hasRadical(numRad):
                        modele._genRadicaux[numRad] = pere.genRadical(numRad)
                    else:
                        warnings.warn(gr + " has no radical {}".format(numRad), MissingRadical)

            # héritage des absents
            modele._absents = pere.absents()

        # génération des désinences suffixées
        ldsuf = []
        clefsSuff = set(list(modele.msuff.keys()))
        for suff in clefsSuff:
            for d in flatten(modele._desinences.values()):
                if d.morphoNum() in modele.msuff[suff]:
                    gq = d.grq()
                    if gq == "-":
                        gq = ""
                    gq += suff
                    dsuf = Desinence(gq, d.morphoNum(), d.numRad(), modele)
                    ldsuf.append(dsuf)

        for dsuf in ldsuf:
            modele._desinences[dsuf.morphoNum()].append(dsuf)
            self.ajDesinence(dsuf)

        return modele

    def _register_lemme(self, lemma):
        """ Register a lemma into the Lemmatiseur

        :param lemma: Lemma to register
        :return:
        """
        if lemma.cle() not in self.lemmatiseur._lemmes:
            self.ajRadicaux(lemma)
            self.lemmatiseur._lemmes[lemma.cle()] = lemma

    def ajExtensions(self):
        """ Lecture du fichier de lemmes étendus """
        self.lisFichierLexique(self.path("lem_ext.la"))

    def ajLexiques(self):
        """ Lecture du fichier de lemmes de base """
        self.lisFichierLexique(self.path("lemmes.la"))

    def ajModeles(self):
        """ Lecture des modèles, et enregistrement de leurs désinences
        """
        sl = []
        lines = [line for line in lignesFichier(self.path("modeles.la"))]
        max = len(lines) - 1
        for i, l in enumerate(lines):
            if l.startswith('$'):
                varname, value = tuple(l.split("="))
                self.lemmatiseur._variables[varname] = value
                continue

            eclats = l.split(":")
            if (eclats[0] == "modele" or i == max) and len(sl) > 0:
                m = self.parse_modele(sl)
                self.register_modele(m)
                sl = []
            sl.append(l)

    def ajRadicaux(self, lemme):
        """ Calcule tous les radicaux du lemme l,
            *  en se servant des modèles, ajoute à ce lemme,
            *  et ensuite à la map *  des radicaux de la classe Lemmat.

        Ligne type de lemme
        # ablŭo=ā̆blŭo|lego|ā̆blŭ|ā̆blūt|is, ere, lui, lutum
        #      0        1    2    3         4

        :param lemme: Lemme
        :type lemme: Lemme
        """
        m = self.lemmatiseur.modele(lemme.grModele())
        ''' insérer d'abord les radicaux définis dans lemmes.la
        qui sont prioritaires '''

        for i in lemme.cles_radicaux():
            radical_list = lemme.radical(i)
            for radical in radical_list:
                self.lemmatiseur._radicaux[deramise(radical.gr()).lower()].append(radical)

        # pour chaque radical du modèle
        for indice_radical in m.cles_radicaux():
            # Si le radical a été défini par le lemme
            if indice_radical in lemme.cles_radicaux():
                continue
            gs = lemme.grq().split(',')
            for graphie in gs:
                gen = m.genRadical(indice_radical)
                # si gen == 'K', radical est la forme canonique
                if gen == "-":
                    continue
                if gen != "K":
                    # sinon, la règle de formation du modèle
                    oter, ajouter = 0, "0"
                    if "," in gen:
                        oter, ajouter = tuple(gen.split(","))
                        oter = int(oter)
                    else:
                        oter = int(gen)

                    if oter == len(graphie):
                        graphie = ""
                    elif oter != 0:
                        graphie = graphie[:-oter]
                    if ajouter != "0":
                        graphie += ajouter
                r = Radical(graphie, indice_radical, lemme)

                # Doute si cela n'appartient pas à graphe in gs
                lemme.ajRadical(indice_radical, r)
                self.lemmatiseur._radicaux[deramise(r.gr()).lower()].append(r)

    def ajDesinence(self, d):
        """ Ajoute la désinence d dans la map des désinences. """
        self.lemmatiseur._desinences[deramise(d.gr())].append(d)

    def parse_irreg(self, l):
        """ Constructeur de la classe Irreg.

        :param l: Ligne de chargement des irréguliers
        :type l: str
        """
        ecl = l.split(':')
        grq = ecl[0]
        exclusif = False
        if grq.endswith("*"):
            grq = grq[:-1]
            exclusif = True
        return Irreg(
            graphie_accentuee=grq, graphie=atone(grq),
            exclusif=exclusif,
            morphos=listeI(ecl[2]),
            lemme=self.lemmatiseur.lemme(ecl[1]),
            parent=self.lemmatiseur
        )



from .ch import atone, listeI


class Irreg(object):

    def __repr__(self):
        return "<pycollatinus.irregs.Irreg[{}]>".format(self._gr)

    def __init__(self, graphie_accentuee: str, graphie: str, lemme, morphos: list,
                 exclusif: bool = False, parent=None):
        """

        :param graphie_accentuee: Graphie accentuee
        :param graphie: Graphie
        :param lemme: Lemma
        :type lemme: pycollatinus.lemme.Lemme
        :param morphos: List of morphologies matched by this form
        :param exclusif: si la forme régulière calculée par le modèle est inusitée, remplace par la forme irrégulière.
        :param parent: Lemmatiseur
        :type parent: pycollatinus.lemmatiseur.Lemmatiseur
        """
        self._lemmat = parent
        self._grq = graphie_accentuee
        self._exclusif = exclusif
        self._gr = graphie
        self._lemme = lemme
        self._morphos = morphos

    def exclusif(self):
        """ True si le lemmes est exclusif, c'est à dire si la forme régulière calculée par le modèle est inusitée, remplace par la forme irrégulière.

        :return: si le lemmes est exclusif
        :rtype: bool
        """
        return self._exclusif

    def gr(self):
        """ Graphie ramiste sans diacritique.

        :return: Graphie ramiste sans diacritique.
        :rtype: str
        """
        return self._gr

    def grq(self):
        """ Graphie ramiste avec diacritiques.

        :return: Graphie ramiset avec diacritiques.
        :rtype: str
        """
        return self._grq

    def lemme(self):
        """ Le lemme de l'irrégulier.

        :return: Le lemme de l'irrégulier.
        :rtype: pycollatinus.lemme.Lemme
        """
        return self._lemme

    def morphos(self):
        """  Liste des numéros de morphos que peut prendre l'irrégulier, en tenant compte des quantités.

        :return: Liste des numéros de morphos que peut prendre l'irrégulier, en tenant compte des quantités.
        :rtype: list of int
        """
        return self._morphos

    def pos(self):
        """ Return the pos of the parent

        :rtype: str
        """
        return self.lemme().pos()

    def cle(self):
        return self.lemme().cle()

    def possible_forms(self):
        """ Generate a list of possible forms for the current lemma

        :returns: List of possible forms for the current lemma
        :rtype: [str]
        """
        return list(set(self.lemme().possible_forms() + [self.gr()]))




import re
from unidecode import unidecode

voyelles = "āăēĕīĭōŏūŭȳўĀĂĒĔĪĬŌŎŪŬȲЎ"
consonnes = "bcdfgjklmnpqrstvxz"
abrev = [
    "Agr", "Ap", "A", "K", "D", "F", "C",
    "Cn", "Kal", "L", "Mam", "M\"", "M", "N", "Oct",
    "Opet", "Post", "Pro", "P", "Q", "Sert",
    "Ser", "Sex", "S", "St", "Ti", "T", "V",
    "Vol", "Vop", "Pl"
]

_Y_SHORT_REPLACE = re.compile("[Ўў]")


_SIMPLIFIED_RE = re.compile("\s+")


def listeI(l):
    """ Fonction importante permettant de renvoyer
            une liste d'entiers à partir d'une chaîne.
            La chaîne est une liste de sections séparées
            par des virgules. Une section peut être soit
            un entier, soit un intervalle d'entiers. On
            donne alors les limites inférieure et supérieure
            de l'intervale, séparées par le caractère '-'.
            Nombreux exemples d'intervalles dans le fichier
            data/modeles.la.

    :param l: Chaîne à transformer
    :type l: str
    :return: Liste des sections étendues
    :rtype: list of int
    """
    result = []
    lvirg = l.split(',')
    for virg in lvirg:
        if "-" in virg:
            deb, fin = tuple(virg.split("-"))
            result += [i for i in range(int(deb), int(fin) + 1)]
        else:
            result.append(int(virg))

    return result


def clean_double_diacritic(string):
    return string.replace("\u014D\u0306", "\u014D").replace("\u0306\u014D", "\u0306")


def simplified(string):
    """ Remove multiple spaces and strip a string

    :param string: Chaîne à transformer
    :type string: str
    :return: Chaîne nettoyée
    :rtype: str
    """
    return _SIMPLIFIED_RE.sub(" ", string).strip()


def atone(string, caps=True):
    """ Supprimer les diacritiques de la forme donnée

    :param string: Chaîne à transformer
    :type string: str
    :param caps: Transforme les majuscules
    :type caps: bool
    :return: Chaîne nettoyée
    :rtype: str
    """
    # Fix for unidecode until unidecode stop replace short y by u
    string = _Y_SHORT_REPLACE.sub("y", string)
    if caps is True:
        return unidecode(string)
    a = ""
    for char in string:
        if char.isupper():
            a += char
        else:
            a += unidecode(char)
    return a


_VOY_COMMUNES = re.compile("\w*[aeiouy]\w*")


_VOY_REPLACE = [
    (re.compile("([^āăō])e"), "\1ē̆"),
    (re.compile("^e"),"ē"),
    (re.compile("([^āēq])u"), "\1ū̆"),
    (re.compile("^u"), "ū̆"),
    (re.compile("^y"),"ȳ̆"),
    (re.compile("([^ā])y"), "\1ȳ̆")
]


def communes(g):
    """ Note comme communes toutes les voyelles qui ne portent pas de quantité.

    :param string: Chaîne à transformer
    :type string: str
    :return: Chaîne nettoyée
    :rtype: str
    """
    if len(g) == 0:
        return g

    maj = g[0].isupper()
    g = g.lower()
    if _VOY_COMMUNES.match(g):
        g = g.replace("a", "ā") \
             .replace("i", "ī") \
             .replace("o", "ō")
        for regex, replace in _VOY_REPLACE:
            g = regex.sub(replace, g)

    if maj:
        g = g[0].upper() + g[1:]

    return g


_ROMAIN_REGEXP = re.compile(r"^M{0,4}(CM|CD|D?C{0,3})(XC|XL|L?X{0,3})(IX|IV|V?I{0,3})$")


def estRomain(f):
    """ F est-elle une forme de nombre romain ?

    :param f: Forme
    :type f: str
    :return: Statut de nombre romain
    :type: bool
    """
    return _ROMAIN_REGEXP.match(f)


def deramise(r):
    """ Déramise une chaîne
    
    :param string: Chaîne à transformer
    :type string: str
    :return: Chaîne nettoyée
    :rtype: str
    """
    return r.replace('J', 'I') \
            .replace('j', 'i') \
            .replace('v', 'u') \
            .replace("æ", "ae") \
            .replace("Æ", "Ae") \
            .replace("œ", "oe") \
            .replace("Œ", "Oe") \
            .replace("ụ", 'u') \
            .replace('V', 'U')


def allonge(f):
    """ Modifie f pour que sa dernière voyelle devienne longue.
    
    :param string: Chaîne à transformer
    :type string: str
    :return: Chaîne modifiée
    :rtype: str
    """
    if not f:
        return ""

    return f

    ## TODO : Implementer correctement la suite

    taille = len(f)
    # Je sais que le morceau à attacher commence par une consonne.
    if f[taille-2] in consonnes and not f[-2:].lower() in ["āe", "āu", "ēu",  "ōe"]:
        f = re.sub("[a\u0103]([" + consonnes + "])$", "\u0101\\1", f)
        f = re.sub("[e\u0115]([" + consonnes + "])$", "\u0113\\1", f)
        f = re.sub("[i\u012d]([" + consonnes + "])$", "\u012b\\1", f)
        f = re.sub("[o\u014F]([" + consonnes + "])$", "\u014d\\1", f)
        f = re.sub("[u\u016d]([" + consonnes + "])$", "\u016b\\1", f)
        f = re.sub("[y\u0233]([" + consonnes + "])$", "\u045e\\1", f)
        f = re.sub("[A\u0102]([" + consonnes + "])$", "\u0100\\1", f)
        f = re.sub("[E\u0114]([" + consonnes + "])$", "\u0112\\1", f)
        f = re.sub("[I\u012c]([" + consonnes + "])$", "\u012a\\1", f)
        f = re.sub("[O\u014e]([" + consonnes + "])$", "\u014c\\1", f)
        f = re.sub("[U\u016c]([" + consonnes + "])$", "\u016a\\1", f)
        f = re.sub("[Y\u0232]([" + consonnes + "])$", "\u040e\\1", f)
        return f


from setuptools import setup, find_packages

version = "0.1.1"

setup(
    name='pycollatinus',
    version=version,
    packages=find_packages(exclude=["tests"]),
    url='https://github.com/ponteineptique/collatinus-python',
    license='GNU GPL v2',
    author='Thibault Clerice, Yves Ouvrard, Philippe Verkerk',
    author_email='leponteineptique@gmail.com',
    description='Collatinus Port for Python',
    test_suite="tests",
    install_requires=[
        "unidecode==0.4.21"
    ],
    test_requires=[
        "coverage==4.4.1"
    ],
    include_package_data=True,
    zip_safe=False
)



class MissingRadical(UserWarning):
    """ Error thrown on missing radical """


class UnknownModeleConfigurationKey(UserWarning):
    """ Error thrown when a configuration key is absolutely unknown """
